unit XButton;

{
  TXButton 1.0 (1999-10-04)
  ============
  TButton descendant with flat, down, hot track and drop down menu support. 

  Freeware.

  Copyright  1999 Roman Stedronsky, Roman.Stedronsky@seznam.cz.

  All rights reserved. You may use this software in an application
  without fee or royalty, provided this copyright notice remains intact.

  properties:

    Alignment: TAlignment	horizontal text alignment (left, right, center)
    Down: boolean		button is pushed down
    DropDownMenu: TPopupMenu	menu appears after clicking or pressing space or tool button
    Flat: boolean		button has no border
    HighlightColor: TColor	color of text, when pointer moves over button
    HotTrack: boolean		enables / disables coloring text
    Layout: TTextLayout		vertical text alignment (top, bottom, center)
    Margin: Integer		extra free space form button edges

 events:

    OnDropDown                  appears after dropping menu
    OnDropping			can disable dropdown menu showing
    OnMouseEnter		appears when mouse cursor enter button
    OnMouseExit			appears when mouse cursor leaves button

}

{$S-,W-,R-}
{$C PRELOAD}

interface

uses
  StdCtrls, Classes, Graphics, Windows, Messages, Controls, Menus;

type
  TXDropDownEvent = procedure(Sender: TObject; X, Y: integer) of object;
  TXDroppingEvent = procedure(Sender: TObject; var AllowDrop: Boolean) of object;

  TXButton = class(TButton)
  private
    FCanvas: TCanvas;
    FMouseInControl: boolean;
    FDroppedDown: boolean;
    IsFocused: Boolean;
    // property variables
    FAlignment: TAlignment;
    FDown: boolean;
    FDropDownMenu: TPopupMenu;
    FFlat: boolean;
    FHighlightColor: TColor;
    FHotTrack: boolean;
    FLayout: TTextLayout;
    FMargin: Integer;
    // event variables
    FOnDropDown: TXDropDownEvent;
    FOnDropping: TXDroppingEvent;
    FOnMouseEnter: TNotifyEvent;
    FOnMouseExit: TNotifyEvent;
    // property manipulation methods
    procedure SetAlignment(Value: TAlignment);
    procedure SetDown(Value: boolean);
    procedure SetDropDownMenu(Value: TPopupMenu);
    procedure SetFlat(Value: boolean);
    procedure SetHighlightColor(Value: TColor);
    procedure SetHotTrack(Value: boolean);
    procedure SetLayout(Value: TTextLayout);
    procedure SetMargin(Value: Integer);
    // message response methods
    procedure CNMeasureItem(var Message: TWMMeasureItem); message CN_MEASUREITEM;
    procedure CNDrawItem(var Message: TWMDrawItem); message CN_DRAWITEM;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMMouseEnter(var msg: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var msg: TMessage); message CM_MOUSELEAVE;
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
  protected
    // core methods
    procedure DrawItem(const DrawItemStruct: TDrawItemStruct);
    procedure DropDown;
    // everrided methods
    procedure CreateParams(var Params: TCreateParams); override;
    procedure SetButtonStyle(ADefault: Boolean); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  public
    // construction / desstruction methods
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    // properties
    property Alignment: TAlignment read FAlignment write SetAlignment default taCenter;
    property Down: boolean read FDown write SetDown default false;
    property DropDownMenu: TPopupMenu read FDropDownMenu write SetDropDownMenu;
    property Flat: boolean read FFlat write SetFlat default false;
    property HighlightColor: TColor read FHighlightColor write SetHighlightColor default clBlue;
    property HotTrack: boolean read FHotTrack write SetHotTrack default false;
    property Layout: TTextLayout read FLayout write SetLayout default tlCenter;
    property Margin: Integer read FMargin write SetMargin default 0;
    // events
    property OnDropDown: TXDropDownEvent read FOnDropDown write FOnDropDown;
    property OnDropping: TXDroppingEvent read FOnDropping write FOnDropping;
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseExit: TNotifyEvent read FOnMouseExit write FOnMouseExit;
  end;

procedure Register;

implementation

uses
  SysUtils;

{ property manipulation methods }

procedure TXButton.SetAlignment(Value: TAlignment);
begin
  if FAlignment <> Value then
  begin
    FAlignment := Value;
    Repaint;
  end;
end;

procedure TXButton.SetDown(Value: boolean);
begin
  if FDown <> Value then
  begin
    FDown := Value;
    Repaint;
  end;
end;

procedure TXButton.SetDropDownMenu(Value: TPopupMenu);
begin
  FDropDownMenu := Value;
  if Value <> nil then
  begin
    Value.ParentBiDiModeChanged(Self);
    Value.FreeNotification(Self);
  end;
end;

procedure TXButton.SetFlat(Value: boolean);
begin
  if FFlat <> Value then
  begin
    FFlat := Value;
    Repaint;
  end;
end;

procedure TXButton.SetHighlightColor(Value: TColor);
begin
  if Value <> FHighlightColor then
  begin
    FHighlightColor := Value;
    Repaint;
  end;
end;

procedure TXButton.SetHotTrack(Value: boolean);
begin
  if Value <> FHotTrack then
  begin
    FHotTrack := Value;
    Repaint;
  end;
end;

procedure TXButton.SetLayout(Value: TTextLayout);
begin
  if Value <> Flayout then
  begin
    FLayout := Value;
    Repaint;
  end;
end;

procedure TXButton.SetMargin(Value: Integer);
begin
  if (Value <> FMargin) and (Value >= -1) then
  begin
    FMargin := Value;
    Invalidate;
  end;
end;

{ message response methods }

procedure TXButton.CNMeasureItem(var Message: TWMMeasureItem);
begin
  with Message.MeasureItemStruct^ do
  begin
    itemWidth := Width;
    itemHeight := Height;
  end;
end;

procedure TXButton.CNDrawItem(var Message: TWMDrawItem);
begin
  DrawItem(Message.DrawItemStruct^);
end;

procedure TXButton.CMFontChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TXButton.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TXButton.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TXButton.CMMouseEnter(var msg: TMessage);
begin
  inherited;
  if {FFlat and } not FMouseInControl and Enabled and (DragMode <> dmAutomatic)
    and (GetCapture = 0) then
  begin
    FMouseInControl := True;
    if (Assigned(FOnMouseEnter)) then
      FOnMouseEnter(Self);
    Repaint;
  end;
end;

procedure TXButton.CMMouseLeave(var msg: TMessage);
begin
  inherited;
  if {FFlat and}  FMouseInControl and Enabled and not Dragging then
  begin
    FMouseInControl := False;
    if (Assigned(FOnMouseExit)) then
      FOnMouseExit(Self);
    Invalidate;
  end;
end;

procedure TXButton.WMLButtonDblClk(var Message: TWMLButtonDblClk);
begin
  Perform(WM_LBUTTONDOWN, Message.Keys, Longint(Message.Pos));
end;

{ core methods }

procedure TXButton.DrawItem(const DrawItemStruct: TDrawItemStruct);
const
  cAlignment: array[TAlignment] of word = (DT_LEFT, DT_RIGHT, DT_CENTER);
  cLayout: array[TTextLayout] of word = (DT_TOP, DT_VCENTER, DT_BOTTOM);
  cFlat: array[boolean] of word = (EDGE_SUNKEN, BDR_SUNKENOUTER);
var
  R: TRect;
  Flags: Longint;
  Delta: integer;
begin
  FCanvas.Handle := DrawItemStruct.hDC;
  R := ClientRect;
  if FFlat then Delta := -2 else Delta := -4;
  // draw frame rect
  if Down then
    DrawEdge(DrawItemStruct.hDC, R, cFlat[Flat], BF_RECT or BF_MIDDLE)
  else if FMouseInControl and FFlat then
    DrawEdge(DrawItemStruct.hDC, R, BDR_RAISEDINNER, BF_RECT or BF_MIDDLE)
  else if FFlat then
  begin
    FCanvas.Brush.Color := Color;
    FCanvas.FillRect(R);
  end
  else
  begin
   if DrawItemStruct.itemState and ODS_FOCUS <> 0 then
    begin
      Rectangle(DrawItemStruct.hDC, R.Left, R.Top, R.Right, R.Bottom);
      InflateRect(R, -1, -1);
      Inc(Delta);
    end;
    Flags := DFCS_BUTTONPUSH;
    if FFlat then Flags := Flags or DFCS_FLAT;
    if DrawItemStruct.itemState and ODS_SELECTED <> 0 then Flags := Flags or DFCS_PUSHED;
    if DrawItemStruct.itemState and ODS_DISABLED <> 0 then Flags := Flags or DFCS_INACTIVE;
    DrawFrameControl(DrawItemStruct.hDC, R, DFC_BUTTON, Flags);
  end;
  // draw focus rect
  InflateRect(R, Delta, Delta);
  if (DrawItemStruct.itemState and ODS_FOCUS <> 0) and not Down then
    DrawFocusRect(DrawItemStruct.hDC, R);
  // draw text
  InflateRect(R, -2 - Margin, -2 - Margin);
  if Down then OffsetRect(R, 1, 1); // text is pushed down & right
  if FMouseInControl and FHotTrack and not FDroppedDown then
    SetTextColor(DrawItemStruct.hDC, ColorToRGB(FHighlightColor))
  else
    SetTextColor(DrawItemStruct.hDC, ColorToRGB(Font.Color));
  SetBkMode(DrawItemStruct.hDC, TRANSPARENT);
  DrawText(DrawItemStruct.hDC, PChar(Caption), Length(Caption), R, cAlignment[Alignment] or cLayout[Layout] or DT_SINGLELINE);
  FCanvas.Handle := 0;
end;

procedure TXButton.DropDown;
var
  Point: TPoint;
  Result: boolean;
  Orig: boolean;
  Pt: TPoint;
  Position: DWORD;
begin
  Result := true;
  if Assigned(FOnDropping) then
    FOnDropping(Self, Result);
  if Result then
  begin
    FDroppedDown := true;
    Orig := Down;
    Down := true;
    SendCancelMode(nil);
    FDropDownMenu.PopupComponent := Self;
    Point.x := Left;
    Point.y := Top + Height;
    Point := Parent.ClientToScreen(Point);
    FDropDownMenu.Popup(Point.x, Point.y);
    Position := GetMessagePos;
    Pt.x := LoWord(Position);
    Pt.y := HiWord(Position);
    FMouseInControl := PtInRect(BoundsRect, Parent.ScreenToClient(Pt));
    FDroppedDown := false;
    if Assigned(FOnDropDown) then
      FOnDropDown(Self, Point.x, Point.y);
    Down := Orig;
    Invalidate;
  end;
end;

{ everrided methods }

procedure TXButton.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
    Style := Style or BS_OWNERDRAW;
end;

procedure TXButton.SetButtonStyle(ADefault: Boolean);
begin
  if ADefault <> IsFocused then
  begin
    IsFocused := ADefault;
    Refresh;
  end;
end;

procedure TXButton.KeyDown(var Key: Word; Shift: TShiftState);
begin
  case Key of
    vk_Apps:
      if Assigned(FDropDownMenu) then DropDown;
    vk_Space:
      begin
        Key := 0;
        if Assigned(FDropDownMenu) then DropDown;
        Exit;
      end;
  end;
  inherited KeyDown(Key, Shift);
end;

procedure TXButton.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if (Button = mbLeft) and Enabled then
    if Assigned(FDropDownMenu) then
      DropDown;
end;

{ construction / destruction methods }

constructor TXButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCanvas := TCanvas.Create;
  FDroppedDown := false;
  ControlStyle := ControlStyle + [csReflector];
  FAlignment := taCenter;
  FDown := false;
  FFlat := false;
  FHighlightColor := clBlue;
  FHotTrack := false;
  FLayout := tlCenter;
  FMargin := 0;
end;

destructor TXButton.Destroy;
begin
  FCanvas.Free;
  inherited Destroy;
end;

{ registration }

procedure Register;
begin
  RegisterComponents('Extra', [TXButton]);
end;

end.

